/* sausage.c -- sausage i/o functions
   by Brennan Underwood, http://brennan.home.ml.org/ */
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>
#ifdef __DJGPP__
#include <unistd.h>
#endif
#ifdef _MSC_VER
#include <io.h>	// open, close, etc.
#include <malloc.h>
#endif

#include "sauspriv.h"
#include "sausage.h"

#ifndef NUM_SAU_ENTRIES
#define NUM_SAU_ENTRIES 512
#endif

//static sausage_index_entry *sindex;
static sausage_index_entry *sindex;
static int num_index_entries=0;
char *sausage_error = NULL;
static int _overlay = 1, _open_standalone = 1;

static int filesize(int fd) {
  struct stat statbuf;
  int r;
#ifdef __DJGPP__
  int sav;
  sav = _djstat_flags;
  /* don't waste time calculating expensive stuff */
  _djstat_flags |= _STAT_INODE | _STAT_EXEC_EXT | _STAT_EXEC_MAGIC;
#endif
  r = fstat(fd, &statbuf);
#ifdef __DJGPP__
  _djstat_flags = sav;
#endif
  if (r != 0) return 0;
  return statbuf.st_size;
}

static sausage_index_entry *search_sausage_index(char *filename) {
  int i;
  if (sindex == NULL) return NULL;
  /* someday we'll binary search this or something */
  for (i = 0; i < num_index_entries; i++)
    if (!strcmp(sindex[i].filename, filename))
      return &sindex[i];
  return NULL;
}

int control_sausage(int control, int param) {
  switch (control) {
    case SAUS_OVERLAY:
      _overlay = param;
    break;
    case SAUS_OPEN_STANDALONE:
      _open_standalone = param;
    break;
  }
  return 1;
}

int mount_sausage(char *filename, int *version) {
  sausage_header head;
  sausage_file_entry entry;
  int entries, fd = -1, i;

  /* open it */
  if ((fd = open(filename, O_RDONLY | O_BINARY)) < 0) {
    sausage_error = "Could not open file";
    return 0;
  }

  /* read the header */
  if (read(fd, &head, sizeof head) != sizeof head) {
    close(fd);
    sausage_error = "Header truncated";
    return 0;
  }

  /* check signature, version */
  if (memcmp(head.signature, "sausage", 7)) {
    close(fd);
    sausage_error = "Not a sausage file";
    return 0;
  }
  if (head.version > SAUSAGE_VERSION) {
    close(fd);
    sausage_error = "Incompatible version";
    return 0;
  }
  /* store the version */
  if (version != NULL) *version = head.version;

  /* how many entries does the sau have */
  entries = head.num_files;

  /* read in the entries */
  for (i = 0; i < entries; i++) {
    sausage_index_entry *iptr;
    if (read(fd, &entry, sizeof entry) != (sizeof entry))
      continue;
    /* see if we can overlay this over another entry */
    iptr = search_sausage_index(entry.filename);
    if (iptr == NULL) {
      /* reallocate the index to hold one more entry */
      sindex = realloc(sindex,
                       (num_index_entries+1) * sizeof(sausage_index_entry));
      iptr = sindex + num_index_entries;
      num_index_entries++;
    } else if (_overlay == 0) {
      continue;
    }
    /* copy the info */
    memset(iptr->filename, 0, 80);
    memset(iptr->sausage, 0, 16);
    strncpy(iptr->filename, entry.filename, 79);
    strncpy(iptr->sausage, filename, 15);
    iptr->offset = entry.offset;
    iptr->len = entry.len;
  }

  sausage_error = NULL;
  close(fd);
  return 1;
}

SAUSAGE_FILE *sausage_fopen(char *filename, char *dummymode) {
  sausage_index_entry *iptr, entry;
  SAUSAGE_FILE *ret;
  char *s, *d, *fixfilename;

  fixfilename = alloca(strlen(filename)+1);

  for (s = filename, d = fixfilename; *s; s++, d++) {
    if (*s == '\\') *d = '/';
    else *d = *s;
  }
  *d = 0;

  ret = calloc(sizeof(SAUSAGE_FILE), 1);
  if (ret == NULL) {
    sausage_error = "Out of memory";
    return NULL;
  }

//fprintf(stderr, "searching for file '%s'\n", fixfilename);
  iptr = search_sausage_index(fixfilename);
  if (iptr == NULL) {
    /* check if we shouldn't open standalone files */
    if (!_open_standalone) return NULL;
    if ((ret->fp = fopen(fixfilename, dummymode)) == NULL) {
      free(ret);
      sausage_error = "File not found";
      return NULL;
    }
    /* found a standalone file, encapsulate it with a mini-sausage */
    fseek(ret->fp, 0, SEEK_SET);
    ret->start = 0;
    ret->end = filesize(fileno(ret->fp));
    sausage_error = NULL;
    return ret;
  }

  entry = *iptr;
  
  if ((ret->fp = fopen(entry.sausage, "rb")) == NULL) {
    sausage_error = "Can't reopen sausage";
    return NULL;
  }
  fseek(ret->fp, entry.offset, SEEK_SET);
  ret->start = entry.offset;
  ret->end = entry.offset + entry.len;
  sausage_error = NULL;
  return ret;
}

void sausage_fclose(SAUSAGE_FILE *sfp) {
  if (sfp == NULL) return;
  if (sfp->fp != NULL) fclose(sfp->fp);
  sfp->fp = NULL;
  free(sfp);
  sausage_error = NULL;
}

int sausage_fread(void *_pt, int item_size, int num_items, SAUSAGE_FILE *sfp) {
  int i, itemsread = 0, r;
  char *pt = _pt;

  if (pt == NULL || sfp == NULL) return 0;

  /* check if we can do it in one block */
  if (item_size * num_items + ftell(sfp->fp) < sfp->end) {
    r = fread(pt, item_size, num_items, sfp->fp);
    return r;
  }

  /* otherwise, do it one item at a time */
  for (i = 0; i < num_items; i++) {
    if (ftell(sfp->fp) + item_size > sfp->end) break;
    r = fread(pt, item_size, 1, sfp->fp);
    if (r == -1) return -1;
    if (r != 1) break;
    pt += item_size;
    itemsread++;
  }

  sausage_error = NULL;
  return itemsread;
}

int sausage_read_int(SAUSAGE_FILE *sfp) {
  int ret=0;
  sausage_fread(&ret, 4, 1, sfp);
  return ret;
}

float sausage_read_float(SAUSAGE_FILE *sfp) {
  float ret=0.0;
  if (sausage_fread(&ret, 4, 1, sfp) != 1) return 0.0;
  return ret;
}

char *sausage_fgets(char *buffer, int maxlen, SAUSAGE_FILE *sfp) {
  int ml, curpos;
#if 1
  *buffer = 0;
  curpos = ftell(sfp->fp);
  if (curpos >= sfp->end) return NULL; /* EOF */
  ml = sfp->end - curpos; /* max chars left */
//cprintf("%d chars left\r\n", ml);
  if (ml < maxlen) maxlen = ml; /* don't go past EOF */
  if (maxlen <= 1) return NULL;
//cprintf("maxlen is %d\r\n", maxlen);
  return fgets(buffer, maxlen, sfp->fp);
#else
  sausage_error = NULL;
  return fgets(buffer, maxlen, sfp->fp);
#endif
}

/* this is still slightly screwed up, but SEEK_SET and _CUR work */
int sausage_fseek(SAUSAGE_FILE *sfp, int seek_type, int offset) {
  int oldpos, newpos;
  int ret;
  oldpos = ftell(sfp->fp);
  switch (seek_type) {
    case SEEK_SET:
      ret = fseek(sfp->fp, seek_type, offset+sfp->start);
    break;
    case SEEK_CUR:
      ret = fseek(sfp->fp, seek_type, offset);
    break;
    case SEEK_END:
      return -1;
    break;
    default:
      return -1;
  }
  if (ret != 0) return ret;
#if 0
  newpos = ftell(sfp->fp);
  if (newpos < sfp->start || newpos > sfp->end) {
    fseek(sfp->fp, SEEK_SET, oldpos); /* put it back */
    return -1;
  }
#endif
  return ret;
}

int sausage_ftell(SAUSAGE_FILE *sfp) {
  return ftell(sfp->fp) - sfp->start;
}
